import { second, shifters } from '../floor-shift-ceil/floor-shift-ceil';
var SPANS_WITH_WEEK = ['year', 'month', 'week', 'day', 'hour', 'minute', 'second'];
var SPANS_WITHOUT_WEEK = ['year', 'month', 'day', 'hour', 'minute', 'second'];
var SPANS_WITHOUT_WEEK_OR_MONTH = ['year', 'day', 'hour', 'minute', 'second'];
var SPANS_UP_TO_DAY = ['day', 'hour', 'minute', 'second'];
function capitalizeFirst(str) {
    if (!str.length)
        return str;
    return str[0].toUpperCase() + str.substr(1);
}
var periodWeekRegExp = /^P(\d+)W$/;
var periodRegExp = /^P(?:(\d+)Y)?(?:(\d+)M)?(?:(\d+)D)?(?:T(?:(\d+)H)?(?:(\d+)M)?(?:(\d+)S)?)?$/;
function getSpansFromString(durationStr) {
    var spans = {};
    var matches;
    if ((matches = periodWeekRegExp.exec(durationStr))) {
        spans.week = Number(matches[1]);
        if (!spans.week)
            throw new Error('Duration can not have empty weeks');
    }
    else if ((matches = periodRegExp.exec(durationStr))) {
        var nums = matches.map(Number);
        for (var i = 0; i < SPANS_WITHOUT_WEEK.length; i++) {
            var span = SPANS_WITHOUT_WEEK[i];
            var value = nums[i + 1];
            if (value)
                spans[span] = value;
        }
    }
    else {
        throw new Error("Can not parse duration '" + durationStr + "'");
    }
    return spans;
}
function getSpansFromStartEnd(start, end, timezone) {
    start = second.floor(start, timezone);
    end = second.floor(end, timezone);
    if (end <= start)
        throw new Error('start must come before end');
    var spans = {};
    var iterator = start;
    for (var i = 0; i < SPANS_WITHOUT_WEEK.length; i++) {
        var span = SPANS_WITHOUT_WEEK[i];
        var spanCount = 0;
        var length_1 = end.valueOf() - iterator.valueOf();
        var canonicalLength = shifters[span].canonicalLength;
        if (length_1 < canonicalLength / 4)
            continue;
        var numberToFit = Math.min(0, Math.floor(length_1 / canonicalLength) - 1);
        var iteratorMove = void 0;
        if (numberToFit > 0) {
            iteratorMove = shifters[span].shift(iterator, timezone, numberToFit);
            if (iteratorMove <= end) {
                spanCount += numberToFit;
                iterator = iteratorMove;
            }
        }
        while (true) {
            iteratorMove = shifters[span].shift(iterator, timezone, 1);
            if (iteratorMove <= end) {
                iterator = iteratorMove;
                spanCount++;
            }
            else {
                break;
            }
        }
        if (spanCount) {
            spans[span] = spanCount;
        }
    }
    return spans;
}
function removeZeros(spans) {
    var newSpans = {};
    for (var i = 0; i < SPANS_WITH_WEEK.length; i++) {
        var span = SPANS_WITH_WEEK[i];
        if (Number(spans[span]) > 0) {
            newSpans[span] = spans[span];
        }
    }
    return newSpans;
}
function fitIntoSpans(length, spansToCheck) {
    var spans = {};
    var lengthLeft = length;
    for (var i = 0; i < spansToCheck.length; i++) {
        var span = spansToCheck[i];
        var spanLength = shifters[span].canonicalLength;
        var count = Math.floor(lengthLeft / spanLength);
        if (count) {
            lengthLeft -= spanLength * count;
            spans[span] = count;
        }
    }
    return spans;
}
var Duration = (function () {
    function Duration(spans, end, timezone) {
        if (spans && end && timezone) {
            spans = getSpansFromStartEnd(spans, end, timezone);
        }
        else if (typeof spans === 'object') {
            spans = removeZeros(spans);
        }
        else {
            throw new Error('new Duration called with bad argument');
        }
        var usedSpans = Object.keys(spans);
        if (!usedSpans.length)
            throw new Error('Duration can not be empty');
        if (usedSpans.length === 1) {
            this.singleSpan = usedSpans[0];
        }
        else if (spans.week) {
            throw new Error("Can not mix 'week' and other spans");
        }
        this.spans = spans;
    }
    Duration.fromJS = function (durationStr) {
        if (typeof durationStr !== 'string')
            throw new TypeError('Duration JS must be a string');
        return new Duration(getSpansFromString(durationStr));
    };
    Duration.fromCanonicalLength = function (length, skipMonths) {
        if (skipMonths === void 0) { skipMonths = false; }
        if (length <= 0)
            throw new Error('length must be positive');
        var spans = fitIntoSpans(length, skipMonths ? SPANS_WITHOUT_WEEK_OR_MONTH : SPANS_WITHOUT_WEEK);
        if (length % shifters['week'].canonicalLength === 0 &&
            (Object.keys(spans).length > 1 ||
                spans['day'])) {
            spans = { week: length / shifters['week'].canonicalLength };
        }
        return new Duration(spans);
    };
    Duration.fromCanonicalLengthUpToDays = function (length) {
        if (length <= 0)
            throw new Error('length must be positive');
        return new Duration(fitIntoSpans(length, SPANS_UP_TO_DAY));
    };
    Duration.prototype.toString = function () {
        var strArr = ['P'];
        var spans = this.spans;
        if (spans.week) {
            strArr.push(String(spans.week), 'W');
        }
        else {
            var addedT = false;
            for (var i = 0; i < SPANS_WITHOUT_WEEK.length; i++) {
                var span = SPANS_WITHOUT_WEEK[i];
                var value = spans[span];
                if (!value)
                    continue;
                if (!addedT && i >= 3) {
                    strArr.push('T');
                    addedT = true;
                }
                strArr.push(String(value), span[0].toUpperCase());
            }
        }
        return strArr.join('');
    };
    Duration.prototype.add = function (duration) {
        return Duration.fromCanonicalLength(this.getCanonicalLength() + duration.getCanonicalLength());
    };
    Duration.prototype.subtract = function (duration) {
        var newCanonicalDuration = this.getCanonicalLength() - duration.getCanonicalLength();
        if (newCanonicalDuration < 0)
            throw new Error('A duration can not be negative.');
        return Duration.fromCanonicalLength(newCanonicalDuration);
    };
    Duration.prototype.multiply = function (multiplier) {
        if (multiplier <= 0)
            throw new Error('Multiplier must be positive non-zero');
        if (multiplier === 1)
            return this;
        var newCanonicalDuration = this.getCanonicalLength() * multiplier;
        return Duration.fromCanonicalLength(newCanonicalDuration);
    };
    Duration.prototype.valueOf = function () {
        return this.spans;
    };
    Duration.prototype.toJS = function () {
        return this.toString();
    };
    Duration.prototype.toJSON = function () {
        return this.toString();
    };
    Duration.prototype.equals = function (other) {
        return other instanceof Duration && this.toString() === other.toString();
    };
    Duration.prototype.isSimple = function () {
        var singleSpan = this.singleSpan;
        if (!singleSpan)
            return false;
        return this.spans[singleSpan] === 1;
    };
    Duration.prototype.isFloorable = function () {
        var singleSpan = this.singleSpan;
        if (!singleSpan)
            return false;
        var span = Number(this.spans[singleSpan]);
        if (span === 1)
            return true;
        var siblings = shifters[singleSpan].siblings;
        if (!siblings)
            return false;
        return siblings % span === 0;
    };
    Duration.prototype.floor = function (date, timezone) {
        var singleSpan = this.singleSpan;
        if (!singleSpan)
            throw new Error('Can not floor on a complex duration');
        var span = this.spans[singleSpan];
        var mover = shifters[singleSpan];
        var dt = mover.floor(date, timezone);
        if (span !== 1) {
            if (!mover.siblings) {
                throw new Error("Can not floor on a ".concat(singleSpan, " duration that is not 1"));
            }
            if (mover.siblings % span !== 0) {
                throw new Error("Can not floor on a ".concat(singleSpan, " duration that does not divide into ").concat(mover.siblings));
            }
            dt = mover.round(dt, span, timezone);
        }
        return dt;
    };
    Duration.prototype.shift = function (date, timezone, step) {
        if (step === void 0) { step = 1; }
        var spans = this.spans;
        for (var _i = 0, SPANS_WITH_WEEK_1 = SPANS_WITH_WEEK; _i < SPANS_WITH_WEEK_1.length; _i++) {
            var span = SPANS_WITH_WEEK_1[_i];
            var value = spans[span];
            if (value)
                date = shifters[span].shift(date, timezone, step * value);
        }
        return date;
    };
    Duration.prototype.materialize = function (start, end, timezone, step) {
        if (step === void 0) { step = 1; }
        var values = [];
        var iter = this.floor(start, timezone);
        while (iter <= end) {
            values.push(iter);
            iter = this.shift(iter, timezone, step);
        }
        return values;
    };
    Duration.prototype.isAligned = function (date, timezone) {
        return this.floor(date, timezone).valueOf() === date.valueOf();
    };
    Duration.prototype.dividesBy = function (smaller) {
        var myCanonicalLength = this.getCanonicalLength();
        var smallerCanonicalLength = smaller.getCanonicalLength();
        return (myCanonicalLength % smallerCanonicalLength === 0 &&
            this.isFloorable() &&
            smaller.isFloorable());
    };
    Duration.prototype.getCanonicalLength = function () {
        var spans = this.spans;
        var length = 0;
        for (var _i = 0, SPANS_WITH_WEEK_2 = SPANS_WITH_WEEK; _i < SPANS_WITH_WEEK_2.length; _i++) {
            var span = SPANS_WITH_WEEK_2[_i];
            var value = spans[span];
            if (value)
                length += value * shifters[span].canonicalLength;
        }
        return length;
    };
    Duration.prototype.getDescription = function (capitalize) {
        var spans = this.spans;
        var description = [];
        for (var _i = 0, SPANS_WITH_WEEK_3 = SPANS_WITH_WEEK; _i < SPANS_WITH_WEEK_3.length; _i++) {
            var span = SPANS_WITH_WEEK_3[_i];
            var value = spans[span];
            var spanTitle = capitalize ? capitalizeFirst(span) : span;
            if (value) {
                if (value === 1) {
                    description.push(spanTitle);
                }
                else {
                    description.push(String(value) + ' ' + spanTitle + 's');
                }
            }
        }
        return description.join(', ');
    };
    Duration.prototype.getSingleSpan = function () {
        return this.singleSpan;
    };
    Duration.prototype.getSingleSpanValue = function () {
        if (!this.singleSpan)
            return;
        return this.spans[this.singleSpan];
    };
    Duration.prototype.limitToDays = function () {
        return Duration.fromCanonicalLengthUpToDays(this.getCanonicalLength());
    };
    return Duration;
}());
export { Duration };
